@@ -15,6 +15,7 @@ from django.db import models
 from django.db.models.query import QuerySet
 from django.urls import reverse
 from django.utils.functional import cached_property
+from math import inf
 from model_utils.managers import InheritanceManager
 import copy
 
@@ -404,30 +405,98 @@ class Text(models.Model):
         return deleted
 
     def save(self, *args, **kwargs):
+
         if self.pk is not None:
             orig = Text.objects.get(pk=self.pk)
             if orig.text != self.text and "apis_highlighter" in settings.INSTALLED_APPS:
-                ann = Annotation.objects.filter(text_id=self.pk).order_by("start")
-                seq = SequenceMatcher(None, orig.text, self.text)
-                for a in ann:
-                    changed = False
-                    count = 0
-                    for s in seq.get_matching_blocks():
-                        count += 1
-                        if s.a <= a.start and (s.a + s.size) >= a.end:
-                            print(a.id)
-                            old_start = copy.deepcopy(a.start)
-                            old_end = copy.deepcopy(a.end)
-                            new_start = a.start + (s.b - s.a)
-                            new_end =  a.end + (s.b - s.a)
-                            if orig.text[old_start:old_end] == self.text[new_start:new_end]:
-                                a.start = new_start
-                                a.end = new_end
-                                a.save()
-                                changed = True
-                                break
-                    if not changed:
-                        a.delete()  # TODO: we might want to delete relations as well.
+                def calculate_context_weights(text, i_start, i_end):
+
+                    def calculate_word_dict(text, direction):
+
+                        word_list = re.split(" |\\n", text)
+                        word_list = [w for w in word_list if w != ""]
+
+                        word_dict = {}
+
+                        if word_list != []:
+
+                            value_step = 1 / len(word_list)
+                            value_current = 1
+                            for word in word_list[::direction]:
+                                word_value = word_dict.get(word, 0)
+                                word_dict[word] = word_value + value_current
+                                value_current -= value_step
+
+                        return word_dict
+
+                    text_left = text[:i_start]
+                    text_right = text[i_end:]
+                    word_dict_left = calculate_word_dict(text=text_left, direction=-1)
+                    word_dict_right = calculate_word_dict(text=text_right, direction=1)
+
+                    return {
+                        "word_dict_left": word_dict_left,
+                        "word_dict_right": word_dict_right,
+                    }
+
+                def make_diff(word_dict_a, word_dict_b):
+
+                    words_all = set(word_dict_a.keys()).union(set(word_dict_b.keys()))
+                    diff_all = 0
+                    for word in words_all:
+                        word_value_a = word_dict_a.get(word, 0)
+                        word_value_b = word_dict_b.get(word, 0)
+                        diff_all += abs(word_value_a - word_value_b)
+
+                    return diff_all
+
+                def correlate_annotations(text_old, text_new, annotations_old):
+
+                    for ann in annotations_old:
+                        i_old_start = ann.start
+                        i_old_end = ann.end
+                        context_weights_dict_old = calculate_context_weights(
+                            text=text_old,
+                            i_start=i_old_start,
+                            i_end=i_old_end
+                        )
+                        ann_text = text_old[ann.start:ann.end]
+                        diff_min = inf
+                        i_new_start = None
+                        i_new_end = None
+                        for i in re.finditer(f"(?={re.escape(ann_text)})", text_new):
+                            i_candidate_start = i.start()
+                            i_candidate_end = i_candidate_start + len(ann_text)
+                            context_weights_dict_new = calculate_context_weights(
+                                text=text_new, i_start=i_candidate_start, i_end=i_candidate_end
+                            )
+                            diff_left = make_diff(
+                                context_weights_dict_new["word_dict_left"],
+                                context_weights_dict_old["word_dict_left"]
+                            )
+                            diff_right = make_diff(
+                                context_weights_dict_new["word_dict_right"],
+                                context_weights_dict_old["word_dict_right"]
+                            )
+                            diff_current = diff_left + diff_right
+                            if diff_current < diff_min:
+                                diff_min = diff_current
+                                i_new_start = i_candidate_start
+                                i_new_end = i_candidate_end
+
+                        if diff_min != inf:
+                            ann.start = i_new_start
+                            ann.end = i_new_end
+                            ann.save()
+                        else:
+                            ann.delete() # TODO: we might want to delete relations as well.
+
+                correlate_annotations(
+                    text_old=orig.text,
+                    text_new=self.text,
+                    annotations_old=Annotation.objects.filter(text_id=self.pk).order_by("start")
+                )
+
         super().save(*args, **kwargs)
 
 
